@@ -361,7 +361,7 @@ doclist.pickDocument = function(name) {
 
 
 var OptionPanel = require("ace/ext/options").OptionPanel;
-var optionsPanel = new OptionPanel(env.editor);
+var optionsPanel = env.optionsPanel = new OptionPanel(env.editor);
 
 var originalAutocompleteCommand = null;
 
@@ -457,7 +457,6 @@ optionsPanel.add({
 var optionsPanelContainer = document.getElementById("optionsPanel");
 optionsPanel.render();
 optionsPanelContainer.insertBefore(optionsPanel.container, optionsPanelContainer.firstChild);
-optionsPanel.container.style.width = "80%";
 optionsPanel.on("setOption", function(e) {
     util.saveOption(e.name, e.value);
 });
@@ -467,6 +466,15 @@ function updateUIEditorOptions() {
     optionsPanel.render();
 }
 
+env.editor.on("changeSession", function() {
+    for (var i in env.editor.session.$options) {
+        var value = util.getOption(i);
+        if (value != undefined) {
+            env.editor.setOption(i, value);
+        }
+    }
+});
+
 optionsPanel.setOption("doc", util.getOption("doc") || "JavaScript");
 for (var i in optionsPanel.options) {
     var value = util.getOption(i);
@@ -490,13 +498,6 @@ function synchroniseScrolling() {
 var StatusBar = require("ace/ext/statusbar").StatusBar;
 new StatusBar(env.editor, cmdLine.container);
 
-
-var Emmet = require("ace/ext/emmet");
-net.loadScript("https://cloud9ide.github.io/emmet-core/emmet.js", function() {
-    Emmet.setCore(window.emmet);
-    env.editor.setOption("enableEmmet", true);
-});
-
 require("ace/placeholder").PlaceHolder;
 
 var snippetManager = require("ace/snippets").snippetManager;
