@@ -37,3 +37,21 @@ String nestedMultilineTemplates() {
         }
         """;
 }
+
+var to = "working";
+
+module com.example.myapp {
+    /**
+     * Test doc comment
+     * @property test
+     */
+    requires transitive java.base;
+    //comment
+    exports com.example.myapp.api;
+    /*
+        multiline comment
+    */
+    opens com.example.myapp.internal to other.module;
+    uses com.example.myapp.service.MyService;
+    provides com.example.myapp.service.MyService with com.example.myapp.impl.MyServiceImpl;
+}
