@@ -3,6 +3,7 @@
  * @typedef {import("./editor").Editor} Editor
  * @typedef {import("../ace-internal").Ace.CompletionProviderOptions} CompletionProviderOptions
  * @typedef {import("../ace-internal").Ace.CompletionOptions} CompletionOptions
+ * @typedef {import("../ace-internal").Ace.Position} Position
  */
 var HashHandler = require("./keyboard/hash_handler").HashHandler;
 var AcePopup = require("./autocomplete/popup").AcePopup;
@@ -27,25 +28,26 @@ var preventParentScroll = require("./lib/scroll").preventParentScroll;
  * @property {string} [docText] - a plain text that would be displayed as an additional popup. If `docHTML` exists,
  * it would be used instead of `docText`.
  * @property {string} [completerId] - the identifier of the completer
+ * @property {boolean} [skipFilter] - a boolean value to decide if the popup item is going to skip the filtering process done using prefix text.
  * @property {import("../ace-internal").Ace.IRange} [range] - An object specifying the range of text to be replaced with the new completion value (experimental)
- * @property {string} [command] - A command to be executed after the completion is inserted (experimental)
+ * @property {any} [command] - A command to be executed after the completion is inserted (experimental)
  * @property {string} [snippet] - a text snippet that would be inserted when the completion is selected
  * @property {string} [value] - The text that would be inserted when selecting this completion.
- * @property {import("../ace-internal").Ace.Completer & {insertMatch:(editor: Editor, data: Completion) => void}} [completer]
+ * @property {import("../ace-internal").Ace.Completer} [completer]
  * @property {boolean} [hideInlinePreview]
  * @export
  */
 
 /**
  * @typedef {BaseCompletion & {snippet: string}} SnippetCompletion
- * @property {string} snippet 
+ * @property {string} snippet
  * @property {string} [value]
  * @export
  */
 
 /**
  * @typedef {BaseCompletion & {value: string}} ValueCompletion
- * @property {string} value 
+ * @property {string} value
  * @property {string} [snippet]
  * @export
  */
@@ -76,12 +78,13 @@ class Autocomplete {
         this.keyboardHandler.bindKeys(this.commands);
         this.parentNode = null;
         this.setSelectOnHover = false;
+        /**@private*/
         this.hasSeen = new Set();
 
         /**
-         *  @property {Boolean} showLoadingState - A boolean indicating whether the loading states of the Autocompletion should be shown to the end-user. If enabled 
+         *  @property {Boolean} showLoadingState - A boolean indicating whether the loading states of the Autocompletion should be shown to the end-user. If enabled
          * it shows a loading indicator on the popup while autocomplete is loading.
-         * 
+         *
          * Experimental: This visualisation is not yet considered stable and might change in the future.
          */
         this.showLoadingState = false;
@@ -128,7 +131,7 @@ class Autocomplete {
 
     $init() {
         /**@type {AcePopup}**/
-        this.popup = new AcePopup(this.parentNode || document.body || document.documentElement); 
+        this.popup = new AcePopup(this.parentNode || document.body || document.documentElement);
         this.popup.on("click", function(e) {
             this.insertMatch();
             e.stop();
@@ -255,6 +258,10 @@ class Autocomplete {
         });
         this.$elements = null;
     }
+
+    /**
+     * @internal
+     */
     onLayoutChange() {
         if (!this.popup.isOpen) return this.unObserveLayoutChanges();
         this.$updatePopupPosition();
@@ -300,7 +307,7 @@ class Autocomplete {
         if (this.popup.tryShow(pos, lineHeight, "top")) {
             return;
         }
-        
+
         this.popup.show(pos, lineHeight);
     }
 
@@ -333,17 +340,17 @@ class Autocomplete {
         }
 
         editor.keyBinding.addKeyboardHandler(this.keyboardHandler);
-        
+
         var newRow;
         if (this.stickySelection)
-            newRow = this.popup.data.indexOf(previousSelectedItem); 
-        if (!newRow || newRow === -1) 
+            newRow = this.popup.data.indexOf(previousSelectedItem);
+        if (!newRow || newRow === -1)
             newRow = 0;
-        
+
         this.popup.setRow(this.autoSelect ? newRow : -1);
-     
+
         // If we stay on the same row, but the content is different, we want to update the popup.
-        if (newRow === oldRow && previousSelectedItem !== this.completions.filtered[newRow]) 
+        if (newRow === oldRow && previousSelectedItem !== this.completions.filtered[newRow])
             this.$onPopupChange();
 
         // If we stay on the same line and have inlinePreview enabled, we want to make sure the
@@ -466,11 +473,11 @@ class Autocomplete {
             this.detach();
         return result;
     }
-    
+
     /**
      * This is the entry point for the autocompletion class, triggers the actions which collect and display suggestions
      * @param {Editor} editor
-     * @param {CompletionOptions} options
+     * @param {CompletionOptions} [options]
      */
     showPopup(editor, options) {
         if (this.editor)
@@ -493,6 +500,11 @@ class Autocomplete {
         this.updateCompletions(false, options);
     }
 
+    /**
+     *
+     * @param {{pos: Position, prefix: string}} [initialPosition]
+     * @return {CompletionProvider}
+     */
     getCompletionProvider(initialPosition) {
         if (!this.completionProvider)
             this.completionProvider = new CompletionProvider(initialPosition);
@@ -510,7 +522,7 @@ class Autocomplete {
 
     /**
      * @param {boolean} keepPopupPosition
-     * @param {CompletionOptions} options
+     * @param {CompletionOptions} [options]
      */
     updateCompletions(keepPopupPosition, options) {
         if (keepPopupPosition && this.base && this.completions) {
@@ -528,7 +540,7 @@ class Autocomplete {
             this.openPopup(this.editor, prefix, keepPopupPosition);
             return;
         }
-        
+
         if (options && options.matches) {
             var pos = this.editor.getSelectionRange().start;
             this.base = this.editor.session.doc.createAnchor(pos.row, pos.column);
@@ -591,7 +603,7 @@ class Autocomplete {
                 }
             // If showLoadingState is true and there is still a completer loading, show 'Loading...'
             // in the top row of the completer popup.
-            this.completions = !finished && this.showLoadingState ? 
+            this.completions = !finished && this.showLoadingState ?
                 new FilteredList(
                     Autocomplete.completionsForLoading.concat(filtered), completions.filterText
                 ) :
@@ -614,12 +626,12 @@ class Autocomplete {
 
     updateDocTooltip() {
         var popup = this.popup;
-        var all = this.completions.filtered;
+        var all = this.completions && this.completions.filtered;
         var selected = all && (all[popup.getHoveredRow()] || all[popup.getRow()]);
         var doc = null;
         if (!selected || !this.editor || !this.popup.isOpen)
             return this.hideDocTooltip();
-        
+
         var completersLength = this.editor.completers.length;
         for (var i = 0; i < completersLength; i++) {
             var completer = this.editor.completers[i];
@@ -668,30 +680,54 @@ class Autocomplete {
 
         var popup = this.popup;
         var rect = popup.container.getBoundingClientRect();
-        tooltipNode.style.top = popup.container.style.top;
-        tooltipNode.style.bottom = popup.container.style.bottom;
 
-        tooltipNode.style.display = "block";
-        if (window.innerWidth - rect.right < 320) {
-            if (rect.left < 320) {
-                if(popup.isTopdown) {
-                    tooltipNode.style.top = rect.bottom + "px";
-                    tooltipNode.style.left = rect.left + "px";
-                    tooltipNode.style.right = "";
-                    tooltipNode.style.bottom = "";
-                } else {
-                    tooltipNode.style.top = popup.container.offsetTop - tooltipNode.offsetHeight + "px";
-                    tooltipNode.style.left = rect.left + "px";
-                    tooltipNode.style.right = "";
-                    tooltipNode.style.bottom = "";
-                }
+        var targetWidth = 400;
+        var targetHeight = 300;
+        var scrollBarSize = popup.renderer.scrollBar.width || 10;
+
+        var leftSize = rect.left;
+        var rightSize = window.innerWidth - rect.right - scrollBarSize;
+        var topSize = popup.isTopdown ? rect.top : window.innerHeight - scrollBarSize - rect.bottom;
+        var scores = [
+            Math.min(rightSize / targetWidth, 1),
+            Math.min(leftSize / targetWidth, 1),
+            Math.min(topSize / targetHeight * 0.9),
+        ];
+        var max = Math.max.apply(Math, scores);
+        var tooltipStyle = tooltipNode.style;
+        tooltipStyle.display = "block";
+
+        if (max == scores[0]) {
+            tooltipStyle.left = (rect.right + 1) + "px";
+            tooltipStyle.right = "";
+            tooltipStyle.maxWidth = targetWidth * max + "px";
+            tooltipStyle.top = rect.top + "px";
+            tooltipStyle.bottom = "";
+            tooltipStyle.maxHeight = Math.min(window.innerHeight - scrollBarSize - rect.top, targetHeight) + "px";
+        } else if (max == scores[1]) {
+            tooltipStyle.right = window.innerWidth - rect.left + "px";
+            tooltipStyle.left = "";
+            tooltipStyle.maxWidth = targetWidth * max + "px";
+            tooltipStyle.top = rect.top + "px";
+            tooltipStyle.bottom = "";
+            tooltipStyle.maxHeight = Math.min(window.innerHeight - scrollBarSize - rect.top, targetHeight) + "px";
+        } else if (max == scores[2]) {
+            tooltipStyle.left = window.innerWidth - rect.left + "px";
+            tooltipStyle.maxWidth = Math.min(targetWidth, window.innerWidth) + "px";
+
+            if (popup.isTopdown) {
+                tooltipStyle.top = rect.bottom + "px";
+                tooltipStyle.left = rect.left + "px";
+                tooltipStyle.right = "";
+                tooltipStyle.bottom = "";
+                tooltipStyle.maxHeight = Math.min(window.innerHeight - scrollBarSize - rect.bottom, targetHeight) + "px";
             } else {
-                tooltipNode.style.right = window.innerWidth - rect.left + "px";
-                tooltipNode.style.left = "";
+                tooltipStyle.top = popup.container.offsetTop - tooltipNode.offsetHeight + "px";
+                tooltipStyle.left = rect.left + "px";
+                tooltipStyle.right = "";
+                tooltipStyle.bottom = "";
+                tooltipStyle.maxHeight = Math.min(popup.container.offsetTop, targetHeight) + "px";
             }
-        } else {
-            tooltipNode.style.left = (rect.right + 1) + "px";
-            tooltipNode.style.right = "";
         }
     }
 
@@ -705,7 +741,11 @@ class Autocomplete {
         if (el.parentNode)
             el.parentNode.removeChild(el);
     }
-    
+
+    /**
+     * @param e
+     * @internal
+     */
     onTooltipClick(e) {
         var a = e.target;
         while (a && a != this.tooltipNode) {
@@ -733,6 +773,30 @@ class Autocomplete {
         this.inlineRenderer = this.popup = this.editor = null;
     }
 
+    /**
+     * @param {Editor} editor
+     * @return {Autocomplete}
+     */
+    static for(editor) {
+        if (editor.completer instanceof Autocomplete) {
+            return editor.completer;
+        }
+        if (editor.completer) {
+            editor.completer.destroy();
+            editor.completer = null;
+        }
+        if (config.get("sharedPopups")) {
+            if (!Autocomplete["$sharedInstance"])
+                Autocomplete["$sharedInstance"] = new Autocomplete();
+            editor.completer = Autocomplete["$sharedInstance"];
+        } else {
+            editor.completer = new Autocomplete();
+            editor.once("destroy", destroyCompleter);
+        }
+        // @ts-expect-error
+        return editor.completer;
+    }
+
 }
 
 Autocomplete.prototype.commands = {
@@ -762,26 +826,6 @@ Autocomplete.prototype.commands = {
     "PageDown": function(editor) { editor.completer.popup.gotoPageDown(); }
 };
 
-
-Autocomplete.for = function(editor) {
-    if (editor.completer instanceof Autocomplete) {
-        return editor.completer;
-    }
-    if (editor.completer) {
-        editor.completer.destroy();
-        editor.completer = null;
-    }
-    if (config.get("sharedPopups")) {
-        if (!Autocomplete["$sharedInstance"])
-            Autocomplete["$sharedInstance"] = new Autocomplete();
-        editor.completer = Autocomplete["$sharedInstance"];
-    } else {
-        editor.completer = new Autocomplete();
-        editor.once("destroy", destroyCompleter);
-    }
-    return editor.completer;
-};
-
 Autocomplete.startCommand = {
     name: "startAutocomplete",
     exec: function(editor, options) {
@@ -800,10 +844,10 @@ Autocomplete.startCommand = {
  * This class is responsible for providing completions and inserting them to the editor
  */
 class CompletionProvider {
-    
+
 
     /**
-     * @param {{pos: import("../ace-internal").Ace.Position, prefix: string}} initialPosition
+     * @param {{pos: Position, prefix: string}} [initialPosition]
      */
     constructor(initialPosition) {
         this.initialPosition = initialPosition;
@@ -840,7 +884,7 @@ class CompletionProvider {
             // TODO add support for options.deleteSuffix
             if (!this.completions)
                 return false;
-            
+
             var replaceBefore = this.completions.filterText.length;
             var replaceAfter = 0;
             if (data.range && data.range.start.row === data.range.end.row) {
@@ -863,7 +907,7 @@ class CompletionProvider {
                     editor.session.remove(range);
                 }
             }
-          
+
             if (data.snippet) {
                 snippetManager.insertSnippet(editor, data.snippet);
             }
@@ -873,7 +917,7 @@ class CompletionProvider {
             if (data.completer && data.completer.onInsert && typeof data.completer.onInsert == "function") {
                 data.completer.onInsert(editor, data);
             }
-            
+
             if (data.command && data.command === "startAutocomplete") {
                 editor.execCommand(data.command);
             }
@@ -898,9 +942,9 @@ class CompletionProvider {
     gatherCompletions(editor, callback) {
         var session = editor.getSession();
         var pos = editor.getCursorPosition();
-    
+
         var prefix = util.getCompletionPrefix(editor);
-    
+
         var matches = [];
         this.completers = editor.completers;
         var total = editor.completers.length;
@@ -974,7 +1018,7 @@ class CompletionProvider {
 
             processResults(results);
         }.bind(this));
-        
+
         isImmediate = false;
         if (immediateResults) {
             var results = immediateResults;
@@ -994,6 +1038,10 @@ class CompletionProvider {
 }
 
 class FilteredList {
+    /**
+     * @param {any} array
+     * @param {string} [filterText]
+     */
     constructor(array, filterText) {
         this.all = array;
         this.filtered = array;
@@ -1001,7 +1049,7 @@ class FilteredList {
         this.exactMatch = false;
         this.ignoreCaption = false;
     }
-    
+
     setFilter(str) {
         if (str.length > this.filterText && str.lastIndexOf(this.filterText, 0) === 0)
             var matches = this.filtered;
@@ -1011,7 +1059,7 @@ class FilteredList {
         this.filterText = str;
         matches = this.filterCompletions(matches, this.filterText);
         matches = matches.sort(function(a, b) {
-            return b.exactMatch - a.exactMatch || b.$score - a.$score 
+            return b.exactMatch - a.exactMatch || b.$score - a.$score
                 || (a.caption || a.value).localeCompare(b.caption || b.value);
         });
 
@@ -1026,12 +1074,17 @@ class FilteredList {
 
         this.filtered = matches;
     }
-    
+
     filterCompletions(items, needle) {
         var results = [];
         var upper = needle.toUpperCase();
         var lower = needle.toLowerCase();
         loop: for (var i = 0, item; item = items[i]; i++) {
+            if (item.skipFilter) {
+                item.$score = item.score;
+                results.push(item);
+                continue;
+            }
             var caption = (!this.ignoreCaption && item.caption) || item.value || item.snippet;
             if (!caption) continue;
             var lastIndex = -1;
