@@ -6,6 +6,7 @@
  * @typedef {import("../ace-internal").Ace.Delta} Delta
  * @typedef {import("../ace-internal").Ace.IRange} IRange
  * @typedef {import("../ace-internal").Ace.SyntaxMode} SyntaxMode
+ * @typedef {import("../ace-internal").Ace.LineWidget} LineWidget
  */
 
 var oop = require("./lib/oop");
@@ -16,6 +17,7 @@ var EventEmitter = require("./lib/event_emitter").EventEmitter;
 var Selection = require("./selection").Selection;
 var TextMode = require("./mode/text").Mode;
 var Range = require("./range").Range;
+var LineWidgets = require("./line_widgets").LineWidgets;
 var Document = require("./document").Document;
 var BackgroundTokenizer = require("./background_tokenizer").BackgroundTokenizer;
 var SearchHighlight = require("./search_highlight").SearchHighlight;
@@ -45,6 +47,8 @@ class EditSession {
         this.$backMarkers = {};
         this.$markerId = 1;
         this.$undoSelect = true;
+        this.$editor = null;
+        this.prevOp = {};
 
         /** @type {FoldLine[]} */
         this.$foldData = [];
@@ -53,10 +57,13 @@ class EditSession {
             return this.join("\n");
         };
 
-        // Set default background tokenizer with Text mode until editor session mode is set 
+        // @experimental
+        this.$gutterCustomWidgets = {};
+
+        // Set default background tokenizer with Text mode until editor session mode is set
         this.bgTokenizer = new BackgroundTokenizer((new TextMode()).getTokenizer(), this);
 
-        
+
         var _self = this;
         this.bgTokenizer.on("update", function(e) {
             _self._signal("tokenizerUpdate", e);
@@ -69,7 +76,12 @@ class EditSession {
             text = new Document(/**@type{string}*/(text));
 
         this.setDocument(text);
+
         this.selection = new Selection(this);
+        this.$onSelectionChange = this.onSelectionChange.bind(this);
+        this.selection.on("changeSelection", this.$onSelectionChange);
+        this.selection.on("changeCursor", this.$onSelectionChange);
+
         this.$bidiHandler = new BidiHandler(this);
 
         config.resetOptions(this);
@@ -77,6 +89,84 @@ class EditSession {
         config._signal("session", this);
 
         this.destroyed = false;
+        this.$initOperationListeners();
+    }
+
+    $initOperationListeners() {
+        /**@type {import("../ace-internal").Ace.Operation | null}*/
+        this.curOp = null;
+        this.on("change", () => {
+            if (!this.curOp) {
+                this.startOperation();
+                this.curOp.selectionBefore = this.$lastSel;
+            }
+            this.curOp.docChanged = true;
+        }, true);
+        this.on("changeSelection", () => {
+            if (!this.curOp) {
+                this.startOperation();
+                this.curOp.selectionBefore = this.$lastSel;
+            }
+            this.curOp.selectionChanged = true;
+        }, true);
+
+        // Fallback mechanism in case current operation doesn't finish more explicitly.
+        // Triggered, for example, when a consumer makes programmatic changes without invoking endOperation afterwards.
+        this.$operationResetTimer = lang.delayedCall(this.endOperation.bind(this, true));
+    }
+
+    /**
+     * Start an Ace operation, which will then batch all the subsequent changes (to either content or selection) under a single atomic operation.
+     * @param {{command?: {name?: string}, args?: any}|undefined} [commandEvent] Optional name for the operation
+     */
+    startOperation(commandEvent) {
+        if (this.curOp) {
+            if (!commandEvent || this.curOp.command) {
+                return;
+            }
+            this.prevOp = this.curOp;
+        }
+        if (!commandEvent) {
+            commandEvent = {};
+        }
+
+        this.$operationResetTimer.schedule();
+        this.curOp = {
+            command: commandEvent.command || {},
+            args: commandEvent.args
+        };
+        this.curOp.selectionBefore = this.selection.toJSON();
+        this._signal("startOperation", commandEvent);
+    }
+
+    /**
+     * End current Ace operation.
+     * Emits "beforeEndOperation" event just before clearing everything, where the current operation can be accessed through `curOp` property.
+     * @param {any} [e]
+     */
+    endOperation(e) {
+        if (this.curOp) {
+            if (e && e.returnValue === false) {
+                this.curOp = null;
+                this._signal("endOperation", e);
+                return;
+            }
+            if (e == true && this.curOp.command && this.curOp.command.name == "mouse") {
+                // When current operation is mousedown, we wait for the mouseup to end the operation.
+                // So during a user selection, we would only end the operation when the final selection is known.
+                return;
+            }
+
+            const currentSelection = this.selection.toJSON();
+            this.curOp.selectionAfter = currentSelection;
+            this.$lastSel = this.selection.toJSON();
+            this.getUndoManager().addSelection(currentSelection);
+
+            this._signal("beforeEndOperation");
+            this.prevOp = this.curOp;
+            this.curOp = null;
+            this._signal("endOperation", e);
+        }
     }
 
     /**
@@ -104,6 +194,35 @@ class EditSession {
         return this.doc;
     }
 
+    /**
+     * Get "widgetManager" from EditSession
+     *
+     * @returns {LineWidgets} object
+     */
+    get widgetManager() {
+        const widgetManager = new LineWidgets(this);
+        // todo remove the widgetManger assignement from lineWidgets constructor when introducing breaking changes
+        this.widgetManager = widgetManager;
+
+        if (this.$editor)
+            widgetManager.attach(this.$editor);
+
+        return widgetManager;
+    }
+
+    /**
+     * Set "widgetManager" in EditSession
+     *
+     * @returns void
+     */
+    set widgetManager(value) {
+        Object.defineProperty(this, "widgetManager", {
+            writable: true,
+            enumerable: true,
+            configurable: true,
+            value: value,
+        });
+    }
     /**
      * @param {Number} docRow The row to work with
      *
@@ -152,14 +271,19 @@ class EditSession {
             this.bgTokenizer.start(0);
     }
 
+    /**
+     * @param e
+     * @internal
+     */
     onChangeFold(e) {
         var fold = e.data;
         this.$resetRowCache(fold.start.row);
     }
 
     /**
-     * 
+     *
      * @param {Delta} delta
+     * @internal
      */
     onChange(delta) {
         this.$modified = true;
@@ -178,7 +302,7 @@ class EditSession {
             }
             this.$undoManager.add(delta, this.mergeUndoDeltas);
             this.mergeUndoDeltas = true;
-            
+
             this.$informUndoManager.schedule();
         }
 
@@ -186,6 +310,10 @@ class EditSession {
         this._signal("change", delta);
     }
 
+    onSelectionChange() {
+        this._signal("changeSelection");
+    }
+
     /**
      * Sets the session text.
      * @param {String} text The new text to place
@@ -213,7 +341,7 @@ class EditSession {
         undoManager.$redoStack = session.history.redo;
         undoManager.mark = session.history.mark;
         undoManager.$rev = session.history.rev;
-    
+
         const editSession = new EditSession(session.value);
         session.folds.forEach(function(fold) {
           editSession.addFold("...", Range.fromPoints(fold.start, fold.end));
@@ -225,10 +353,10 @@ class EditSession {
         editSession.setScrollTop(session.scrollTop);
         editSession.setUndoManager(undoManager);
         editSession.selection.fromJSON(session.selection);
-    
+
         return editSession;
     }
- 
+
     /**
      * Returns the current edit session.
      * @method toJSON
@@ -249,7 +377,7 @@ class EditSession {
             value: this.doc.getValue()
         };
     }
- 
+
     /**
      * Returns the current [[Document `Document`]] as a string.
      * @method toString
@@ -322,10 +450,10 @@ class EditSession {
      **/
     setUndoManager(undoManager) {
         this.$undoManager = undoManager;
-        
+
         if (this.$informUndoManager)
             this.$informUndoManager.cancel();
-        
+
         if (undoManager) {
             var self = this;
             undoManager.addSession(this);
@@ -375,7 +503,7 @@ class EditSession {
     setUseSoftTabs(val) {
         this.setOption("useSoftTabs", val);
     }
-    
+
     /**
      * Returns `true` if soft tabs are being used, `false` otherwise.
      * @returns {Boolean}
@@ -460,6 +588,34 @@ class EditSession {
         this._signal("changeBreakpoint", {});
     }
 
+    /**
+     * Replaces the custom icon with the fold widget if present from a specific row in the gutter
+     * @param {number} row The row number for which to hide the custom icon
+     * @experimental
+     */
+    removeGutterCustomWidget(row) {
+        if(this.$editor) {
+            this.$editor.renderer.$gutterLayer.$removeCustomWidget(row);
+        }
+    }
+
+    /**
+     * Replaces the fold widget if present with the custom icon from a specific row in the gutter
+     * @param {number} row - The row number where the widget will be displayed
+     * @param {Object} attributes - Configuration attributes for the widget
+     * @param {string} attributes.className - CSS class name for styling the widget
+     * @param {string} attributes.label - Text label to display in the widget
+     * @param {string} attributes.title - Tooltip text for the widget
+     * @param {Object} attributes.callbacks - Event callback functions for the widget e.g onClick; 
+     * @returns {void}
+     * @experimental
+    */
+    addGutterCustomWidget(row,attributes) {
+        if(this.$editor) {
+            this.$editor.renderer.$gutterLayer.$addCustomWidget(row,attributes);
+        }
+    }
+
     /**
      * Removes `className` from the `row`.
      * @param {Number} row The row number
@@ -774,13 +930,14 @@ class EditSession {
 
     /**
      * Reloads all the tokens on the current session. This function calls [[BackgroundTokenizer.start `BackgroundTokenizer.start ()`]] to all the rows; it also emits the `'tokenizerUpdate'` event.
+     * @internal
      **/
     onReloadTokenizer(e) {
         var rows = e.data;
         this.bgTokenizer.start(rows.first);
         this._signal("tokenizerUpdate", e);
     }
-    
+
     /**
      * Sets a new text mode for the `EditSession`. This method also emits the `'changeMode'` event. If a [[BackgroundTokenizer `BackgroundTokenizer`]] is set, the `'tokenizerUpdate'` event is also emitted.
      * @param {SyntaxMode | string} mode Set a new text mode
@@ -808,6 +965,9 @@ class EditSession {
         // load on demand
         this.$modeId = path;
         config.loadModule(["mode", path], function(m) {
+            if (this.destroyed) {
+                return;
+            }
             if (this.$modeId !== path)
                 return cb && cb();
             if (this.$modes[path] && !options) {
@@ -835,9 +995,9 @@ class EditSession {
     $onChangeMode(mode, $isPlaceholder) {
         if (!$isPlaceholder)
             this.$modeId = mode.$id;
-        if (this.$mode === mode) 
+        if (this.$mode === mode)
             return;
-            
+
         var oldMode = this.$mode;
         this.$mode = mode;
 
@@ -861,7 +1021,7 @@ class EditSession {
         /**@type {RegExp}*/
         this.nonTokenRe = mode.nonTokenRe;
 
-        
+
         if (!$isPlaceholder) {
             // experimental method, used by c9 findiniles
             if (mode.attachToSession)
@@ -902,7 +1062,7 @@ class EditSession {
      * @param {Number} scrollTop The new scroll top value
      **/
     setScrollTop(scrollTop) {
-        // TODO: should we force integer lineheight instead? scrollTop = Math.round(scrollTop); 
+        // TODO: should we force integer lineheight instead? scrollTop = Math.round(scrollTop);
         if (this.$scrollTop === scrollTop || isNaN(scrollTop))
             return;
 
@@ -945,7 +1105,7 @@ class EditSession {
      **/
     getScreenWidth() {
         this.$computeWidth();
-        if (this.lineWidgets) 
+        if (this.lineWidgets)
             return Math.max(this.getLineWidgetMaxWidth(), this.screenWidth);
         return this.screenWidth;
     }
@@ -1057,7 +1217,7 @@ class EditSession {
     remove(range) {
         return this.doc.remove(range);
     }
-    
+
     /**
      * Removes a range of full lines. This method also triggers the `'change'` event.
      * @param {Number} firstRow The first row to be removed
@@ -1074,7 +1234,7 @@ class EditSession {
     /**
      * Reverts previous changes to your document.
      * @param {Delta[]} deltas An array of previous changes
-     * @param {Boolean} [dontSelect] [If `true`, doesn't select the range of where the change occured]{: #dontSelect}
+     * @param {Boolean} [dontSelect] If `true`, doesn't select the range of where the change occured
      **/
     undoChanges(deltas, dontSelect) {
         if (!deltas.length)
@@ -1131,14 +1291,14 @@ class EditSession {
     /**
      * Enables or disables highlighting of the range where an undo occurred.
      * @param {Boolean} enable If `true`, selects the range of the reinserted change
-     *      
+     *
      **/
     setUndoSelect(enable) {
         this.$undoSelect = enable;
     }
 
     /**
-     * 
+     *
      * @param {Delta[]} deltas
      * @param {boolean} [isUndo]
      * @return {Range}
@@ -1161,7 +1321,7 @@ class EditSession {
                 }
                 continue;
             }
-            
+
             if (isInsert(delta)) {
                 point = delta.start;
                 if (range.compare(point.row, point.column) == -1) {
@@ -1294,7 +1454,7 @@ class EditSession {
     }
 
     /**
-     * 
+     *
      * @param {number} firstRow
      * @param {number} lastRow
      * @param [dir]
@@ -1324,7 +1484,7 @@ class EditSession {
             x.end.row += diff;
             return x;
         });
-        
+
         var lines = dir == 0
             ? this.doc.getLines(firstRow, lastRow)
             : this.doc.removeFullLines(firstRow, lastRow);
@@ -1434,7 +1594,7 @@ class EditSession {
         }
         return range;
     }
-    
+
     /**
      * Sets whether or not line wrapping is enabled. If `useWrapMode` is different than the current value, the `'changeWrapMode'` event is emitted.
      * @param {Boolean} useWrapMode Enable (or disable) wrap mode
@@ -1510,7 +1670,7 @@ class EditSession {
     }
 
     /**
-     * 
+     *
      * @param {number} wrapLimit
      * @param {number} [min]
      * @param {number} [max]
@@ -1533,7 +1693,7 @@ class EditSession {
     getWrapLimit() {
         return this.$wrapLimit;
     }
-    
+
     /**
      * Sets the line length for soft wrap in the editor. Lines will break
      *  at a minimum of the given length minus 20 chars and at a maximum
@@ -1543,7 +1703,7 @@ class EditSession {
     setWrapLimit(limit) {
         this.setWrapLimitRange(limit, limit);
     }
-    
+
     /**
      * Returns an object that defines the minimum and maximum of the wrap limit; it looks something like this:
      *
@@ -1571,7 +1731,7 @@ class EditSession {
         var lastRow = end.row;
         var len = lastRow - firstRow;
         var removedFolds = null;
-        
+
         this.$updating = true;
         if (len != 0) {
             if (action === "remove") {
@@ -1966,13 +2126,13 @@ class EditSession {
         var h = 1;
         if (this.lineWidgets)
             h += this.lineWidgets[row] && this.lineWidgets[row].rowCount || 0;
-        
+
         if (!this.$useWrapMode || !this.$wrapData[row])
             return h;
         else
             return this.$wrapData[row].length + h;
     }
-    
+
     /**
      * @param {Number} row
      * @returns {Number}
@@ -2264,7 +2424,7 @@ class EditSession {
                 wrapIndent = screenRowOffset > 0 ? wrapRow.indent : 0;
             }
         }
-        
+
         if (this.lineWidgets && this.lineWidgets[row] && this.lineWidgets[row].rowsAbove)
             screenRow += this.lineWidgets[row].rowsAbove;
 
@@ -2347,7 +2507,7 @@ class EditSession {
             if (!maxScreenColumn)
                 maxScreenColumn = Infinity;
             screenColumn = screenColumn || 0;
-            
+
             var c, column;
             for (column = 0; column < str.length; column++) {
                 c = str.charAt(column);
@@ -2361,7 +2521,7 @@ class EditSession {
                     break;
                 }
             }
-            
+
             return [screenColumn, column];
         };
     }
@@ -2386,11 +2546,16 @@ class EditSession {
             this.bgTokenizer.cleanup();
             this.destroyed = true;
         }
+        this.endOperation();
         this.$stopWorker();
         this.removeAllListeners();
         if (this.doc) {
             this.doc.off("change", this.$onChange);
         }
+        if (this.selection) {
+            this.selection.off("changeCursor", this.$onSelectionChange);
+            this.selection.off("changeSelection", this.$onSelectionChange);
+        }
         this.selection.detach();
     }
 }
@@ -2430,7 +2595,7 @@ EditSession.prototype.$wrapLimitRange = {
     max : null
 };
 /**
- * 
+ *
  * @type {null | import("../ace-internal").Ace.LineWidget[]}
  */
 EditSession.prototype.lineWidgets = null;
@@ -2527,7 +2692,7 @@ config.defineOptions(EditSession.prototype, "session", {
             return "off";
         },
         handlesSet: true
-    },    
+    },
     wrapMethod: {
         /**
          * @param {"code"|"text"|"auto"|boolean} val
@@ -2557,7 +2722,7 @@ config.defineOptions(EditSession.prototype, "session", {
                 this.setUseWrapMode(true);
             }
         },
-        initialValue: true 
+        initialValue: true
     },
     firstLineNumber: {
         set: function() {this._signal("changeBreakpoint");},
@@ -2617,4 +2782,3 @@ config.defineOptions(EditSession.prototype, "session", {
 });
 
 exports.EditSession = EditSession;
-
