@@ -22,7 +22,7 @@ var valueResetRegex = isIOS ? /\s/ : /\n/;
 var isMobile = useragent.isMobile;
 
 var TextInput;
-TextInput= function(parentNode, host) {
+TextInput= function(/**@type{HTMLTextAreaElement} */parentNode, /**@type{import("../editor").Editor} */host) {
     /**@type {HTMLTextAreaElement & {msGetInputContext?: () => {compositionStartOffset: number}, getInputContext?: () => {compositionStartOffset: number}}}*/
     var text = dom.createElement("textarea");
     text.className = "ace_text-input";
@@ -35,7 +35,7 @@ TextInput= function(parentNode, host) {
     text.style.opacity = "0";
     parentNode.insertBefore(text, parentNode.firstChild);
 
-    var copied = false;
+    /**@type{boolean|string}*/var copied = false;
     var pasted = false;
     /**@type {(boolean|Object) & {context?: any, useTextareaForIME?: boolean, selectionStart?: number, markerRange?: any}}} */
     var inComposition = false;
@@ -62,7 +62,7 @@ TextInput= function(parentNode, host) {
 
     // Set number of extra lines in textarea, some screenreaders
     // perform better with extra lines above and below in the textarea.
-    this.setNumberOfExtraLines = function(number) {
+    this.setNumberOfExtraLines = function(/**@type{number}*/number) {
         rowStart = Number.MAX_SAFE_INTEGER;
         rowEnd =  Number.MIN_SAFE_INTEGER;
 
@@ -73,6 +73,19 @@ TextInput= function(parentNode, host) {
         
         numberOfExtraLines = number;
     };
+
+    this.setAriaLabel = function() {
+        var ariaLabel = "";
+        if (host.$textInputAriaLabel) {
+            ariaLabel += `${host.$textInputAriaLabel}, `;
+        }
+        if(host.session) {
+            var row = host.session.selection.cursor.row;
+            ariaLabel += nls("text-input.aria-label", "Cursor at row $0", [row + 1]);
+        }
+        text.setAttribute("aria-label", ariaLabel);
+    };
+
     this.setAriaOptions = function(options) {
         if (options.activeDescendant) {
             text.setAttribute("aria-haspopup", "true");
@@ -88,19 +101,11 @@ TextInput= function(parentNode, host) {
         }     
         if (options.setLabel) {
             text.setAttribute("aria-roledescription", nls("text-input.aria-roledescription", "editor"));
-            var arialLabel = "";
-            if (host.$textInputAriaLabel) {
-                arialLabel += `${host.$textInputAriaLabel}, `;
-            }
-            if(host.session) {
-                var row =  host.session.selection.cursor.row;
-                arialLabel += nls("text-input.aria-label", "Cursor at row $0", [row + 1]);
-            }
-            text.setAttribute("aria-label", arialLabel);
+            this.setAriaLabel();
         }
     };
 
-    this.setAriaOptions({role: "textbox"}); 
+    this.setAriaOptions({role: "textbox"});
 
     event.addListener(text, "blur", function(e) {
         if (ignoreFocusEvents) return;
@@ -191,6 +196,9 @@ TextInput= function(parentNode, host) {
         // sync value of textarea
         resetSelection();
     });
+
+    // if cursor changes position, we need to update the label with the correct row
+    host.on("changeSelection", this.setAriaLabel);
     
     // Convert from row,column position to the linear position with respect to the current
     // block of lines in the textarea.
@@ -687,7 +695,7 @@ TextInput= function(parentNode, host) {
         var rect = host.container.getBoundingClientRect();
         var style = dom.computedStyle(host.container);
         var top = rect.top + (parseInt(style.borderTopWidth) || 0);
-        var left = rect.left + (parseInt(rect.borderLeftWidth) || 0);
+        var left = rect.left + (parseInt(style.borderLeftWidth) || 0);
         var maxTop = rect.bottom - top - text.clientHeight -2;
         var move = function(e) {
             dom.translate(text, e.clientX - left - 2, Math.min(e.clientY - top - 2, maxTop));
