@@ -86,14 +86,19 @@ class Gutter{
             }
 
             var className = annotation.className;
-            if (className) 
+            if (className) {
                 rowInfo.className = className;
-            else if (annoType == "error")
+            } else if (annoType === "error") {
                 rowInfo.className = " ace_error";
-            else if (annoType == "warning" && rowInfo.className != " ace_error")
+            } else if (annoType === "security" && !/\bace_error\b/.test(rowInfo.className)) {
+                rowInfo.className = " ace_security";
+            } else if (annoType === "warning" && !/\bace_(error|security)\b/.test(rowInfo.className)) {
                 rowInfo.className = " ace_warning";
-            else if (annoType == "info" && (!rowInfo.className))
+            } else if (annoType === "info" && !rowInfo.className) {
                 rowInfo.className = " ace_info";
+            } else if (annoType === "hint" && !rowInfo.className) {
+                rowInfo.className = " ace_hint";
+            }
         }
     }
 
@@ -327,6 +332,7 @@ class Gutter{
         var textNode = element.childNodes[0];
         var foldWidget = element.childNodes[1];
         var annotationNode = element.childNodes[2];
+        var customWidget = element.childNodes[3];
         var annotationIconNode = annotationNode.firstChild;
 
         var firstLineNumber = session.$firstLineNumber;
@@ -374,24 +380,29 @@ class Gutter{
         if (c) {
             var foldClass = "ace_fold-widget ace_" + c;
             var isClosedFold = c == "start" && row == foldStart && row < fold.end.row;
-            if (isClosedFold){
+            if (isClosedFold) {
                 foldClass += " ace_closed";
-                var foldAnnotationClass = '';
+                var foldAnnotationClass = "";
                 var annotationInFold = false;
 
-                for (var i = row + 1; i <= fold.end.row; i++){
-                    if (!this.$annotations[i])
-                        continue;
+                for (var i = row + 1; i <= fold.end.row; i++) {
+                    if (!this.$annotations[i]) continue;
 
-                    if (this.$annotations[i].className === " ace_error"){
+                    if (this.$annotations[i].className === " ace_error") {
                         annotationInFold = true;
                         foldAnnotationClass = " ace_error_fold";
                         break;
-                    } 
-                    if (this.$annotations[i].className === " ace_warning"){
+                    }
+
+                    if (this.$annotations[i].className === " ace_security") {
+                        annotationInFold = true;
+                        foldAnnotationClass = " ace_security_fold";
+                    } else if (
+                        this.$annotations[i].className === " ace_warning" &&
+                        foldAnnotationClass !== " ace_security_fold"
+                    ) {
                         annotationInFold = true;
                         foldAnnotationClass = " ace_warning_fold";
-                        continue;
                     }
                 }
 
@@ -404,27 +415,42 @@ class Gutter{
 
             dom.setStyle(foldWidget.style, "height", lineHeight);
             dom.setStyle(foldWidget.style, "display", "inline-block");
-            
+
             // Set a11y properties.
             foldWidget.setAttribute("role", "button");
             foldWidget.setAttribute("tabindex", "-1");
+
             var foldRange = session.getFoldWidgetRange(row);
 
             // getFoldWidgetRange is optional to be implemented by fold modes, if not available we fall-back.
             if (foldRange)
-                foldWidget.setAttribute("aria-label", nls("gutter.code-folding.range.aria-label", "Toggle code folding, rows $0 through $1", [foldRange.start.row + 1, foldRange.end.row + 1]));
+                foldWidget.setAttribute(
+                    "aria-label",
+                    nls("gutter.code-folding.range.aria-label", "Toggle code folding, rows $0 through $1", [
+                        foldRange.start.row + 1,
+                        foldRange.end.row + 1
+                    ])
+                );
             else {
                 if (fold)
-                    foldWidget.setAttribute("aria-label", nls("gutter.code-folding.closed.aria-label", "Toggle code folding, rows $0 through $1", [fold.start.row + 1, fold.end.row + 1]));
+                    foldWidget.setAttribute(
+                        "aria-label",
+                        nls("gutter.code-folding.closed.aria-label", "Toggle code folding, rows $0 through $1", [
+                            fold.start.row + 1,
+                            fold.end.row + 1
+                        ])
+                    );
                 else
-                    foldWidget.setAttribute("aria-label", nls("gutter.code-folding.open.aria-label", "Toggle code folding, row $0", [row + 1]));
+                    foldWidget.setAttribute(
+                        "aria-label",
+                        nls("gutter.code-folding.open.aria-label", "Toggle code folding, row $0", [row + 1])
+                    );
             }
 
             if (isClosedFold) {
                 foldWidget.setAttribute("aria-expanded", "false");
                 foldWidget.setAttribute("title", nls("gutter.code-folding.closed.title", "Unfold code"));
-            }
-            else {
+            } else {
                 foldWidget.setAttribute("aria-expanded", "true");
                 foldWidget.setAttribute("title", nls("gutter.code-folding.open.title", "Fold code"));
             }
@@ -436,6 +462,14 @@ class Gutter{
                 foldWidget.removeAttribute("aria-label");
             }
         }
+        // fold logic ends here 
+        const customWidgetAttributes = this.session.$gutterCustomWidgets[row];
+        if (customWidgetAttributes) {
+            this.$addCustomWidget(row, customWidgetAttributes,cell);
+        }
+        else if (customWidget){
+            this.$removeCustomWidget(row,cell);
+        }
 
         if (annotationInFold && this.$showFoldedAnnotations){
             annotationNode.className = "ace_gutter_annotation";
@@ -448,11 +482,15 @@ class Gutter{
             var ariaLabel;
             switch(foldAnnotationClass) {
                 case " ace_error_fold":
-                    ariaLabel = nls("gutter.annotation.aria-label.error", "Read annotations row $0", [rowText]);
+                    ariaLabel = nls("gutter.annotation.aria-label.error", "Error, read annotations row $0", [rowText]);
+                    break;
+
+                case " ace_security_fold":
+                    ariaLabel = nls("gutter.annotation.aria-label.security", "Security finding, read annotations row $0", [rowText]);
                     break;
 
                 case " ace_warning_fold":
-                    ariaLabel = nls("gutter.annotation.aria-label.warning", "Read annotations row $0", [rowText]);
+                    ariaLabel = nls("gutter.annotation.aria-label.warning", "Warning, read annotations row $0", [rowText]);
                     break;
             }
             annotationNode.setAttribute("aria-label", ariaLabel);
@@ -474,15 +512,23 @@ class Gutter{
             var ariaLabel;
             switch(this.$annotations[row].className) {
                 case " ace_error":
-                    ariaLabel = nls("gutter.annotation.aria-label.error", "Read annotations row $0", [rowText]);
+                    ariaLabel = nls("gutter.annotation.aria-label.error", "Error, read annotations row $0", [rowText]);
+                    break;
+
+                case " ace_security":
+                    ariaLabel = nls("gutter.annotation.aria-label.security", "Security finding, read annotations row $0", [rowText]);
                     break;
 
                 case " ace_warning":
-                    ariaLabel = nls("gutter.annotation.aria-label.warning", "Read annotations row $0", [rowText]);
+                    ariaLabel = nls("gutter.annotation.aria-label.warning", "Warning, read annotations row $0", [rowText]);
                     break;
 
                 case " ace_info":
-                    ariaLabel = nls("gutter.annotation.aria-label.info", "Read annotations row $0", [rowText]);
+                    ariaLabel = nls("gutter.annotation.aria-label.info", "Info, read annotations row $0", [rowText]);
+                    break;
+
+                case " ace_hint":
+                    ariaLabel = nls("gutter.annotation.aria-label.hint", "Suggestion, read annotations row $0", [rowText]);
                     break;
             }
             annotationNode.setAttribute("aria-label", ariaLabel);
@@ -507,7 +553,7 @@ class Gutter{
         cell.text = rowText;
 
         // If there are no annotations or fold widgets in the gutter cell, hide it from assistive tech.
-        if (annotationNode.style.display === "none" && foldWidget.style.display === "none")
+        if (annotationNode.style.display === "none" && foldWidget.style.display === "none" && !customWidgetAttributes)
             cell.element.setAttribute("aria-hidden", true);
         else
             cell.element.setAttribute("aria-hidden", false);
@@ -552,6 +598,118 @@ class Gutter{
     getShowFoldWidgets() {
         return this.$showFoldWidgets;
     }
+    
+    /**
+     * Hides the fold widget/icon from a specific row in the gutter
+     * @param {number} row The row number from which to hide the fold icon
+     * @param {any} cell - Gutter cell 
+     * @experimental
+     */
+    $hideFoldWidget(row, cell) {
+        const rowCell = cell || this.$getGutterCell(row);
+        if (rowCell && rowCell.element) {
+            const foldWidget = rowCell.element.childNodes[1];
+            if (foldWidget) {
+                dom.setStyle(foldWidget.style, "display", "none");
+            }
+        }
+    }
+
+    /**
+     * Shows the fold widget/icon from a specific row in the gutter
+     * @param {number} row The row number from which to show the fold icon
+     * @param {any} cell - Gutter cell 
+     * @experimental
+     */
+    $showFoldWidget(row,cell) {
+        const rowCell = cell || this.$getGutterCell(row);
+        if (rowCell && rowCell.element) {
+            const foldWidget = rowCell.element.childNodes[1];
+            if (foldWidget && this.session.foldWidgets && this.session.foldWidgets[rowCell.row]) {
+                dom.setStyle(foldWidget.style, "display", "inline-block");
+            }
+        }
+    }
+
+    /**
+    * Retrieves the gutter cell element at the specified cursor row position.
+    * @param {number} row - The row number in the editor where the gutter cell is located starts from 0
+    * @returns {HTMLElement|null} The gutter cell element at the specified row, or null if not found
+    * @experimental
+    */
+    $getGutterCell(row) {
+        // contains only visible rows
+        const cells = this.$lines.cells;
+        const visibileRow= this.session.documentToScreenRow(row,0);
+        // subtracting the first visible screen row index and folded rows from the row number.
+        return cells[row - this.config.firstRowScreen - (row-visibileRow)];
+    }
+
+    /**
+    * Displays a custom widget for a specific row
+    * @param {number} row - The row number where the widget will be displayed
+    * @param {Object} attributes - Configuration attributes for the widget
+    * @param {string} attributes.className - CSS class name for styling the widget
+    * @param {string} attributes.label - Text label to display in the widget
+    * @param {string} attributes.title - Tooltip text for the widget
+    * @param {Object} attributes.callbacks - Event callback functions for the widget e.g onClick; 
+    * @param {any} cell - Gutter cell 
+    * @returns {void}
+    * @experimental
+    */
+    $addCustomWidget(row, {className, label, title, callbacks}, cell) {
+        this.session.$gutterCustomWidgets[row] = {className, label, title, callbacks};
+        this.$hideFoldWidget(row,cell);
+
+        // cell is required because when cached cell is used to render, $lines won't have that cell
+        const rowCell = cell || this.$getGutterCell(row);
+        if (rowCell && rowCell.element) {
+            let customWidget = rowCell.element.querySelector(".ace_custom-widget");
+            // deleting the old custom widget to remove the old click event listener
+            if (customWidget) {
+                customWidget.remove();
+            }
+
+            customWidget = dom.createElement("span");
+            customWidget.className = `ace_custom-widget ${className}`;
+            customWidget.setAttribute("tabindex", "-1");
+            customWidget.setAttribute("role", 'button');
+            customWidget.setAttribute("aria-label", label);
+            customWidget.setAttribute("title", title);
+            dom.setStyle(customWidget.style, "display", "inline-block");
+            dom.setStyle(customWidget.style, "height", "inherit");
+            
+            if (callbacks&& callbacks.onClick) {
+                customWidget.addEventListener("click", (e) => {
+                    callbacks.onClick(e, row);
+                    e.stopPropagation();
+                });
+            }
+
+            rowCell.element.appendChild(customWidget);
+        }
+    }
+    
+    /**
+    * Remove a custom widget for a specific row
+    * @param {number} row - The row number where the widget will be removed
+    * @param {any} cell - Gutter cell 
+    * @returns {void}
+    * @experimental
+    */
+    $removeCustomWidget(row, cell) {
+        delete this.session.$gutterCustomWidgets[row];
+        this.$showFoldWidget(row,cell);
+
+        // cell is required because when cached cell is used to render, $lines won't have that cell
+        const rowCell = cell || this.$getGutterCell(row);
+        if (rowCell && rowCell.element) {
+            const customWidget = rowCell.element.querySelector(".ace_custom-widget");
+            if (customWidget) {
+                rowCell.element.removeChild(customWidget);
+            }
+        }
+    }
 
     $computePadding() {
         if (!this.element.firstChild)
