@@ -134,6 +134,56 @@ module.exports = {
         });
     },
 
+
+    "test scrollbars after value change": function() {
+        editor.container.style.height = "0px";
+        editor.setOptions({
+            maxLines: 8,
+        });
+        var renderCount = 0;
+        editor.renderer.on("afterRender", function(e) {
+            renderCount++;
+        });
+        // horizontal scroll
+        editor.setValue("\n");
+        editor.resize(true);
+        editor.renderer.$loop._flush(); // 1
+        assert.notOk(editor.renderer.scrollBarH.isVisible);
+        assert.notOk(editor.renderer.scrollBar.isVisible);
+        editor.setValue("\n\n\n\n" + "_".repeat(30));
+        editor.resize(true);
+        editor.renderer.$loop._flush(); // 2
+        assert.notOk(editor.renderer.scrollBarH.isVisible);
+        assert.notOk(editor.renderer.scrollBar.isVisible);
+        editor.setValue("\n\n\n\n" + "_".repeat(100));
+        editor.resize(true);
+        editor.renderer.$loop._flush(); // 3
+        assert.ok(editor.renderer.scrollBarH.isVisible);
+        assert.notOk(editor.renderer.scrollBar.isVisible);
+        // vertical scroll
+        editor.setValue("\n".repeat(9));
+        editor.resize(true);
+        editor.renderer.$loop._flush(); // 4
+        assert.notOk(editor.renderer.scrollBarH.isVisible);
+        assert.ok(editor.renderer.scrollBar.isVisible);
+        // vertical and horizontal scroll
+        editor.setValue("\n");
+        editor.resize(true);
+        editor.renderer.$loop._flush(); // 5
+        editor.setValue("\n".repeat(9) + "_".repeat(100));
+        editor.resize(true);
+        editor.renderer.$loop._flush(); // 6
+        assert.notOk(editor.renderer.scrollBarH.isVisible);
+        assert.ok(editor.renderer.scrollBar.isVisible);
+        editor.resize(true);
+        editor.renderer.$loop._flush(); // 7
+        // autosize changes vscroll value in which case updates longestLine
+        // this is why it renders an extra time
+        assert.ok(editor.renderer.scrollBarH.isVisible);
+        assert.ok(editor.renderer.scrollBar.isVisible);
+        assert.equal(renderCount, 7);
+    },
+
     "test autosize from 0 height": function() {
         editor.container.style.height = "0px";
         editor.textInput.getElement().style.position = "fixed";
@@ -239,10 +289,11 @@ module.exports = {
         editor._signal("input", {});
         assert.equal(editor.renderer.content.textContent, "only visible for empty value");
     },
-    "test: highlight indent guide": function () {
+    "test: highlight indent guide": function (done) {
         editor.session.setValue(
             "function Test() {\n" + "    function Inner() {\n" + "        \n" + "        \n" + "    }\n" + "}");
         editor.setOption("highlightIndentGuides", false);
+        editor.setOption("wrap", 10); // to make sure higlight works with wrapped lines
         editor.session.selection.$setSelection(1, 22, 1, 22);
         editor.resize(true);
 
@@ -260,6 +311,14 @@ module.exports = {
         editor.session.selection.$setSelection(1, 15, 1, 15);
         editor.resize(true);
         assertIndentGuides( 0);
+
+        editor.session.selection.clearSelection();
+        editor.session.selection.$setSelection(4, 5, 4, 5);
+
+        setTimeout(() => {
+            assertIndentGuides( 2);
+            done();
+        }, 100);
     },
     "test annotation marks": function() {
         function findPointFillStyle(imageData, x, y) {
